/*
 * BottomUpHierarchicalAligner.cpp
 *
 *  Created on: 17.01.2013
 *      Author: kaeshammer
 */


#include "BottomUpHierarchicalAligner.h"

#include <iostream>


namespace LCFRS {

bool debug = false;

std::ostream& operator<<(std::ostream& os, const MonoItem& item)
{
  os << "[" << item.l << ", ";
  for (boost::dynamic_bitset<>::size_type i = 0; i < item.bs.size(); ++i) {
    os << item.bs[i];
  }
  os << "]";
  return os;
}

std::string bitsetToString(boost_bitset bs) {
  std::string res;
  for (boost::dynamic_bitset<>::size_type i = 0; i < bs.size(); ++i) {
    if (bs.test(i)) {
      res.append("1");
    }
    else {
      res.append("0");
    }
  }
  return res;
}

std::string MonoItem::toString() {
  std::string res;
  res.append("[");
  res.append(l);
  res.append(", ");
  res.append(bitsetToString(bs));
  res.append("]");
  return res;
}

std::string BiItem::toString() {
  std::string res;
  res.append("<");
  res.append(itemSource.toString());
  res.append(", ");
  res.append(itemTarget.toString());
  res.append(">");
  return res;
}

std::ostream& operator<<(std::ostream& os, const BiItem& item)
{
  os << "<" << item.itemSource << ", " << item.itemTarget << ">";
  return os;
}


std::ostream& operator<<(std::ostream& os, const BiBitset& bs)
{
  for (boost_bitset::size_type i = 0; i < (bs.first).size(); ++i) {
    os << bs.first[i];
  }
  os << ", ";
  for (boost_bitset::size_type i = 0; i < (bs.second).size(); ++i) {
    os << bs.second[i];
  }
  return os;
}

bool combine(const boost_bitset& sourceA, const boost_bitset& sourceB,
        const boost_bitset& targetA, const boost_bitset& targetB, BiBitset& b)
{
  if (sourceA.intersects(sourceB) || targetA.intersects(targetB)) {
    return false;
  }
  b = std::make_pair(sourceA|sourceB,targetA|targetB);
  return true;
}

bool combine(const boost_bitset& sourceA, const boost_bitset& sourceB, const boost_bitset& sourceC,
        const boost_bitset& targetA, const boost_bitset& targetB, const boost_bitset& targetC,
        BiBitset& b)
{
  if (sourceA.intersects(sourceB) || targetA.intersects(targetB)
      || sourceB.intersects(sourceC) || targetB.intersects(targetC)
      || sourceC.intersects(sourceA) || targetC.intersects(targetA)) {
    return false;
  }
  boost_bitset sourceAB = sourceA|sourceB;
  boost_bitset targetAB = targetA|targetB;
  b = std::make_pair(sourceAB|sourceC,targetAB|targetC);
  return true;
}

unsigned int numBlocks(const boost_bitset& bs)
{
  unsigned int numBlocks = 0;
  size_t index = bs.find_first();
  int last = -2;
  while (index != boost_bitset::npos) {
    if ((int)index > last+1) {
      numBlocks++;
    }
    last = index;
    index = bs.find_next(index);
  }
  return numBlocks;
}

const std::string BottomUpHierarchicalAligner::NONTERM = "X";
const std::string BottomUpHierarchicalAligner::PRETERM = "T";

bool BottomUpHierarchicalAligner::success()
{
  return goals.size() > 0;
}


void BottomUpHierarchicalAligner::printChart()
{
	std::cout << m_chart << std::endl;
}

void BottomUpHierarchicalAligner::printTrace(BiItem* item, int indent)
{
  for (int i = 0; i < indent; i++) {
	  std::cout << "  ";
  }
  std::cout << *item << std::endl;
  if (item->back1 != NULL) {
    printTrace(item->back1, indent+1);
  }
  if (item->back2 != NULL) {
    printTrace(item->back2, indent+1);
  }
  if (item->back3 != NULL) {
    printTrace(item->back3, indent+1);
  }
}

// TODO
//int BottomUpHierarchicalAligner::numGoalItems()
//{
//  return goals.size();
//}
//
//void BottomUpHierarchicalAligner::traceAsStringHelper(BiItem* item, int indent, std::string& curr)
//{
//  //std::string copy = curr;
//  //for (std::vector<BiItem*>::iterator iter = item->equiItems.begin();
//  //    iter != item ->equiItems.end(); iter++) {
//  //  traceAsString(*iter, indent, copy);
//  //}
//
//
//  for (int i = 0; i < indent; i++) {
//    curr.append("  ");
//  }
//  curr.append(item->toString());
//  curr.append("\n");
//  if (item->back1 != NULL) {
//    traceAsStringHelper(item->back1, indent+1, curr);
//  }
//  if (item->back2 != NULL) {
//    traceAsStringHelper(item->back2, indent+1, curr);
//  }
//
//  //std::cout << "copy: " << copy << std::endl;
//}
//
//void BottomUpHierarchicalAligner::traceAsString(BiItem* item, int indent, std::string& curr)
//{
//  traceAsStringHelper(item, indent, curr);
//  std::cout << curr << std::endl;
//}
//
//void BottomUpHierarchicalAligner::printGoalItems()
//{
//  for (std::vector<BiItem*>::iterator iter = goals.begin();
//  	  iter != goals.end(); iter++) {
//    std::string curr;
//    traceAsString(*iter, 0, curr);
//  }
//}

void BottomUpHierarchicalAligner::printOneDerivation()
{
  if (goals.size() == 0) return;

  printTrace(goals[0], 0);
}


bool BottomUpHierarchicalAligner::validateHierAlignment()
{
  // identify unaligned words
  std::vector<int> sourceUnaligned;
  std::vector<int> targetUnaligned;
  for (std::vector<TranslationUnit>::const_iterator it = m_tuStore.getTUs().begin();
        it != m_tuStore.getTUs().end();
        it++)
  {
    TranslationUnit tu = *it;
    if (tu.isOnlySource()) {
      for (std::set<int>::const_iterator it2 = tu.getSourceIndeces().begin();
              it2 != tu.getSourceIndeces().end(); it2++) {
        sourceUnaligned.push_back(*it2);
      }
    }
    if (tu.isOnlyTarget()) {
      for (std::set<int>::const_iterator it2 = tu.getTargetIndeces().begin();
              it2 != tu.getTargetIndeces().end(); it2++) {
        targetUnaligned.push_back(*it2);
      }
    }
  }
  // sort them
  std::sort(sourceUnaligned.begin(), sourceUnaligned.end());
  std::sort(targetUnaligned.begin(), targetUnaligned.end());

  // build mappings (-> unaligned words)
  // not needed anymore, but we leave it in for now, who knows...
  unsigned int j = 0;
  for (unsigned int i = 0; i < m_tuStore.numSource; i++) {
    if(std::find(sourceUnaligned.begin(), sourceUnaligned.end(), i) == sourceUnaligned.end()) {
      // i is aligned
      in2outSource[i] = j;
      out2inSource[j] = i;
      j++;
    }
  }
  j = 0;
  for (unsigned int i = 0; i < m_tuStore.numTarget; i++) {
    if(std::find(targetUnaligned.begin(), targetUnaligned.end(), i) == targetUnaligned.end()) {
      // i is aligned
      in2outTarget[i] = j;
      out2inTarget[j] = i;
      j++;
    }
  }

  // initial item for each TU
  // push on agenda
  std::vector<BiItem*> transporter;
  for (std::vector<TranslationUnit>::const_iterator it = m_tuStore.getTUs().begin();
      it != m_tuStore.getTUs().end();
      it++)
  {
    TranslationUnit tu = *it;
    if (tu.isOnlySource() || tu.isOnlyTarget()) {
      // we will take care of this separately for efficiency reasons
      continue;
    }
    // construct bitsets
    boost::dynamic_bitset<> bsSource(m_tuStore.numSource);//bsSource(out2inSource.size());
    boost::dynamic_bitset<> bsTarget(m_tuStore.numTarget);//bsTarget(out2inTarget.size());
    for (std::set<int>::const_iterator it2 = tu.getSourceIndeces().begin();
        it2 != tu.getSourceIndeces().end(); it2++)
    {
        bsSource.set(*it2);
    }
    for (std::set<int>::const_iterator it3 = tu.getTargetIndeces().begin();
        it3 != tu.getTargetIndeces().end(); it3++)
    {
        bsTarget.set(*it3);
    }
    //std::cout << (*it).toString() << std::endl;
    //std::cout << bsSource << "   " << bsTarget << std::endl;


    // construct item
    BiItem* item = new BiItem(PRETERM, bsSource, PRETERM, bsTarget);
    //std::cout << *item << std::endl;
    transporter.push_back(item);
  }

  // treat unaligned words
  unsigned int firstI = 0;
  for (unsigned int i = 0; i < sourceUnaligned.size(); i++) {
    if (i+1 == sourceUnaligned.size() || sourceUnaligned[i]+1 != sourceUnaligned[i+1]) {
      // build new item from firstI to i
      boost_bitset bsSource(m_tuStore.numSource);
      boost_bitset bsTarget(m_tuStore.numTarget);
      for (unsigned int j = firstI; j <= i; j++) {
        bsSource.set(sourceUnaligned[j]);
      }
      BiItem* item = new BiItem(PRETERM, bsSource, PRETERM, bsTarget);
      transporter.push_back(item);
      // update
      firstI = i+1;
    }
  }
  firstI = 0;
  for (unsigned int i = 0; i < targetUnaligned.size(); i++) {
    if (i+1 == targetUnaligned.size() || targetUnaligned[i]+1 != targetUnaligned[i+1]) {
      // build new item from firstI to i
      boost_bitset bsSource(m_tuStore.numSource);
      boost_bitset bsTarget(m_tuStore.numTarget);
      for (unsigned int j = firstI; j <= i; j++) {
        bsTarget.set(targetUnaligned[j]);
      }
      BiItem* item = new BiItem(PRETERM, bsSource, PRETERM, bsTarget);
      transporter.push_back(item);
      // update
      firstI = i+1;
    }
  }

  // for the new items, put them on the agenda
  for (std::vector<BiItem*>::iterator it = transporter.begin();
      it != transporter.end(); it++) {
    //std::cout << **it << std::endl;
    agenda.push_back(*it);
  }
  transporter.clear();


//  for (std::list<BiItem*>::iterator it = agenda.begin(); it != agenda.end(); it++) {
//    std::cout << *(*it) << std::endl;
//  }

  unsigned int m = !(m_maxNumBlocksSource < m_maxNumBlocksTarget) ?
      m_maxNumBlocksTarget : m_maxNumBlocksSource;

  for (unsigned int i = 1; i <= m; i++) {
    if (doParse(i,i)) {
      return true;
    }
    else {
      //std::cout << "another try after " << i << "|" << i << std::endl;
      // copy all items from chart to agenda, but mark them as special
      // then run next loop of doParse
      for (my_chart::const_iterator iter = m_chart.chart.begin();
          iter != m_chart.chart.end(); iter++) {
        // inner chart
        for (my_inner_chart::const_iterator iter2 = iter->second.begin();
            iter2 != iter->second.end(); iter2++) {
          iter2->second->status = true;
          agenda.push_back(iter2->second);
        }
      }
    }
  }

  if (m_maxNumBlocksSource == m_maxNumBlocksTarget) {
    return false;
  }
  else if (m_maxNumBlocksSource > m_maxNumBlocksTarget) {
    for (unsigned int i = m+1; i <= m_maxNumBlocksSource; i++) {
      if (doParse(i,m)) {
        return true;
      }
      else {
        std::cout << "another try after " << i << "|" << m << std::endl;
        for (my_chart::const_iterator iter = m_chart.chart.begin();
            iter != m_chart.chart.end(); iter++) {
          for (my_inner_chart::const_iterator iter2 = iter->second.begin();
              iter2 != iter->second.end(); iter2++) {
            iter2->second->status = true;
            agenda.push_back(iter2->second);
          }
        }
      }
    }
    return false;
  }
  else if (m_maxNumBlocksSource < m_maxNumBlocksTarget) {
    for (unsigned int i = m+1; i <= m_maxNumBlocksTarget; i++) {
      if (doParse(m,i)) {
        return true;
      }
      else {
        std::cout << "another try after " << m << "|" << i << std::endl;
        for (my_chart::const_iterator iter = m_chart.chart.begin();
            iter != m_chart.chart.end(); iter++) {
          for (my_inner_chart::const_iterator iter2 = iter->second.begin();
              iter2 != iter->second.end(); iter2++) {
            iter2->second->status = true;
            agenda.push_back(iter2->second);
          }
        }
      }
    }
    return false;
  }
  return false; // should never be reached
}

bool BottomUpHierarchicalAligner::doParse(unsigned int kS, unsigned int kT)
{
  std::vector<BiItem*> transporter;

  // while agenda is not empty
  while (!agenda.empty()) {
    // pull first item
    BiItem* item = agenda.front();
    agenda.pop_front();

    if (item->status == false) {

      // put item in the chart
      bool alreadyIn = false;
      m_chart.addItem(item, alreadyIn);
      if (alreadyIn) {
        continue;
      }
    }

    if (debug) std::cout << "pop: " << *item << " " << item << std::endl;

    // check whether it is a goal item
    if (item->itemSource.l == NONTERM
        && item->itemTarget.l == NONTERM
        && item->itemSource.bs.count() == m_tuStore.numSource
        && item->itemTarget.bs.count() == m_tuStore.numTarget) {
      goals.push_back(item);
      break;
    }


    // combine item with each item in the chart
    // check for the new item whether it fulfills the constraints
    // if yes, put it on the agenda

    if (item->itemSource.l == PRETERM && item->itemTarget.l == PRETERM) { // T-item

      // UNARY
      // < X -> T, X -> T >
      if (numBlocks(item->itemSource.bs) <= kS
          && numBlocks(item->itemTarget.bs) <= kT) {
        // constraints met
        BiItem* nitem = new BiItem(NONTERM, item->itemSource.bs, NONTERM, item->itemTarget.bs, item);
        transporter.push_back(nitem);
      }

      // < X -> T[1] X[2], X -> T[1] X[2] >
      // and
      // < X -> T[1] X[2] X[3], X -> T[1] X[2] X[3] >
      if (!m_normalForm) { // only allowed if noNF

        // find X-items
        my_chart::const_iterator found = m_chart.get(std::make_pair(NONTERM,NONTERM));
        if (found != m_chart.chart.end()) {
          my_inner_chart chart2 = (*found).second;
          // for each of them, check whether the bitset combination is valid
          // UNARYMIXED
          for(my_inner_chart::iterator iter = chart2.begin(); iter != chart2.end(); iter++)
          {
            if (debug) std::cout << "  combine with " << iter->first << std::endl;
            BiBitset combbs;
            if (combine(iter->first.first, item->itemSource.bs,
                iter->first.second, item->itemTarget.bs, combbs)) {
              if (numBlocks(combbs.first) <= kS
                  && numBlocks(combbs.second) <= kT) {
                BiItem* nitem = new BiItem(NONTERM, combbs.first, NONTERM, combbs.second, iter->second, item);
                if (debug) std::cout << "      -> " << *nitem << std::endl;
                transporter.push_back(nitem);
              }
            }

            // BINARYMIXED
            for(my_inner_chart::iterator iter2 = chart2.begin(); iter2 != chart2.end(); iter2++) {
              BiBitset combbs2;
              if (combine(iter->first.first, item->itemSource.bs, iter2->first.first,
                  iter->first.second, item->itemTarget.bs, iter2->first.second, combbs2)) {
                if (numBlocks(combbs2.first) <= kS
                    && numBlocks(combbs2.second) <= kT) {
                  BiItem* nitem = new BiItem(NONTERM, combbs2.first, NONTERM, combbs2.second,
                      iter2->second, iter->second, item);
                  //std::cout << "!!! Binary Mixed item " << *nitem << std::endl;
                  if (debug) std::cout << "      -> " << *nitem << std::endl;
                  transporter.push_back(nitem);
                }
              }
            }
          }
        }
      }
    }

    if (item->itemSource.l == NONTERM && item->itemTarget.l == NONTERM) { // X-item
      // find X-items
      my_chart::const_iterator foundX = m_chart.get(std::make_pair(NONTERM,NONTERM));
      // find T-items
      my_chart::const_iterator foundT = m_chart.get(std::make_pair(PRETERM,PRETERM));

      // < X -> X[1] X[2], X -> X[1] X[2] >
      // and
      // < X -> T[1] X[2] X[3], X -> T[1] X[2] X[3] >
      if (foundX != m_chart.chart.end()) {
        my_inner_chart chart2 = (*foundX).second;
        // for each of them, check whether the bitset combination is valid
        for(my_inner_chart::iterator iter = chart2.begin(); iter != chart2.end(); iter++)
        {
          // BINARY
          //if (debug) std::cout << "  combine with " << iter->first << std::endl;
          BiBitset combbs;
          if (combine(iter->first.first, item->itemSource.bs,
                               iter->first.second, item->itemTarget.bs, combbs)) {
            if (numBlocks(combbs.first) <= kS
                && numBlocks(combbs.second) <= kT) {
              BiItem* nitem = new BiItem(NONTERM, combbs.first, NONTERM, combbs.second, iter->second, item);
              if (debug) std::cout << "      -> " << *nitem << nitem << std::endl;
              transporter.push_back(nitem);
            }
          }
          // BINARYMIXED
          if (!m_normalForm && foundT != m_chart.chart.end()) {
            my_inner_chart chartT = (*foundT).second;
            for (my_inner_chart::iterator iterT = chartT.begin(); iterT != chartT.end(); iterT++) {
              //if (debug) std::cout << "         combine with " << iterT->first << std::endl;
              BiBitset combbs2;
              if (combine(iter->first.first, item->itemSource.bs, iterT->first.first,
                  iter->first.second, item->itemTarget.bs, iterT->first.second, combbs2)) {
                if (numBlocks(combbs2.first) <= kS
                    && numBlocks(combbs2.second) <= kT) {
                  BiItem* nitem = new BiItem(NONTERM, combbs2.first, NONTERM, combbs2.second,
                      iterT->second, iter->second, item);
                  //std::cout << "!!! Binary Mixed item " << *nitem << std::endl;
                  if (debug) std::cout << "      -> " << *nitem << nitem << std::endl;
                  transporter.push_back(nitem);
                }
              }
            }
          }
        }
      }

      // UNARYMIXED
      // < X -> T[1] X[2], X -> T[1] X[2] >
      // find T-items
      if (!m_normalForm) { // only allowed if noNF
        if (foundT != m_chart.chart.end()) {
          my_inner_chart chart2 = (*foundT).second;
          // for each of them, check whether the bitset combination is valid
          for(my_inner_chart::iterator iter = chart2.begin(); iter != chart2.end(); iter++)
          {
            //if (debug) std::cout << "  combine with " << iter->first << std::endl;
            BiBitset combbs;
            if (combine(iter->first.first, item->itemSource.bs,
                                 iter->first.second, item->itemTarget.bs, combbs)) {
              if (numBlocks(combbs.first) <= kS
                  && numBlocks(combbs.second) <= kT) {
                BiItem* nitem = new BiItem(NONTERM, combbs.first, NONTERM, combbs.second, iter->second, item);
                if (debug) std::cout << "      -> " << *nitem << nitem << std::endl;
                transporter.push_back(nitem);
              }
            }
          }
        }
      }
    }

    // for the new items, but them on the agenda and in the chart
    for (std::vector<BiItem*>::iterator it = transporter.begin();
        it != transporter.end(); it++) {
      //std::cout << **it << std::endl;
        agenda.push_back(*it);
        //m_chart.addItem(*it);
    }
    transporter.clear();
  }


  // return (there is an item in the chart that covers the whole input)
  return goals.size() > 0;
}

BottomUpHierarchicalAligner::~BottomUpHierarchicalAligner()
{
  // delete all the items in the chart. we had built them with "new".
  for (my_chart::iterator iter = m_chart.chart.begin();
      iter != m_chart.chart.end();
      iter++) {
    for (my_inner_chart::iterator iter2 = iter->second.begin();
        iter2 != iter->second.end();
        iter2++) {
      for (std::vector<BiItem*>::iterator iter3 = iter2->second->equiItems.begin();
          iter3 != iter2->second->equiItems.end();
          iter3++) {
        delete *iter3;
      }
      delete iter2->second;
    }
  }
}

void BiChart::addItem(BiItem* item)
{
  bool in = false;
  addItem(item, in);
}

void BiChart::addItem(BiItem* item, bool& alreadyIn)
{
  BiLabel labels(item->itemSource.l, item->itemTarget.l);
  BiBitset bitsets(item->itemSource.bs, item->itemTarget.bs);

  my_chart::iterator found = chart.find(labels);
  if (found != chart.end()) {
    // labels were found
    my_inner_chart chart2 = (*found).second;
    my_inner_chart::iterator found2 = chart2.find(bitsets);
    if (found2 != chart2.end()) {
      // bitsets are already in there, i.e. an equi-item exists
      found2->second->equiItems.push_back(item);
      alreadyIn = true;
      //old chart2[bitsets].push_back(item);
    }
    else {
      chart2[bitsets] = item;
      alreadyIn = false;
    }
    chart[labels] = chart2; //?
  }
  else {
    // labels are not yet in the chart
    my_inner_chart chart2;
    chart2[bitsets] = item;
    chart[labels] = chart2;
    alreadyIn = false;
  }
}

bool BiChart::equiItemExists(BiItem* item)
{
  BiLabel labels(item->itemSource.l, item->itemTarget.l);
  BiBitset bitsets(item->itemSource.bs, item->itemTarget.bs);
  my_chart::iterator found = chart.find(labels);
  if (found != chart.end()) {
    // labels were found
    my_inner_chart::iterator found2 = found->second.find(bitsets);
    if (found2 != found->second.end()) {
      // bitsets found
      return true;
    }
  }
  return false;
}

my_chart::const_iterator BiChart::get(const BiLabel& label)
{
  return chart.find(label);
}

std::ostream& operator<<(std::ostream& os, const BiChart& bichart)
{
  for (my_chart::const_iterator iter = bichart.chart.begin();
      iter != bichart.chart.end(); iter++) {
    os << "<" << (iter->first).first << ", " << (iter->first).second << ">" << std::endl;

    // inner chart
    my_inner_chart chart2 = iter->second;
    for (my_inner_chart::const_iterator iter2 = chart2.begin();
    		iter2 != chart2.end(); iter2++) {

    	// item
    	os << "     " << *(iter2->second) << "       ";

    	//list
    	std::vector<BiItem*> list = iter2->second->equiItems;
    	unsigned int k = list.size() < 5 ? list.size() : 5;
    	for (unsigned int i = 0; i < k; i++) {
    		os << *list[i] << ", ";
    	}
    	if (list.size() > k) {
    	  os << "...";
    	}
    	os << std::endl;
    }
  }
  return os;
}


}

