/*
 * ExtractLCFRS.cpp
 *
 *  Created on: 09.01.2013
 *      Author: kaeshammer
 */

#include "ExtractLCFRS.h"

#include "Alignment.h"
#include "InputFileStream.h"
#include "Options.h"
#include "OutputFileStream.h"
#include "TranslationUnitStore.h"
#include "BottomUpHierarchicalAligner.h"

#include <boost/program_options.hpp>


namespace LCFRS {


int ExtractLCFRS::Main(int argc, char *argv[])
{
  // Process command-line options.
  Options options;
  ProcessOptions(argc, argv, options);

  // Open input files.
  Moses::InputFileStream targetStream(options.targetFile);
  Moses::InputFileStream sourceStream(options.sourceFile);
  Moses::InputFileStream alignmentStream(options.alignmentFile);

  // Open output files.
  Moses::OutputFileStream outputStream;
  Moses::OutputFileStream tuStream;
  std::string outputFileName = options.extractFile;
  //std::string tuFileName = options.extractFile + std::string(".tu");

  OpenOutputFileOrDie(outputFileName, outputStream);
  //OpenOutputFileOrDie(tuFileName, tuStream);

  std::vector<int> sentNumbers;
  if (!options.sentNumFile.empty()) {
    // only look at specific sentences
    Moses::InputFileStream sentNumberStream(options.sentNumFile);
    std::string line;
    while (true) {
      std::getline(sentNumberStream, line);
      if (sentNumberStream.eof()) {
        break;
      }
      sentNumbers.push_back(boost::lexical_cast<int>(line));
    }
  }

  // print output header
  PrintOutputHeader(outputStream, options.maxNumBlocksSource, options.maxNumBlocksTarget,
      options.maxLengthSource, options.maxLengthTarget, options.normalForm);

  // Reading input files
  std::string targetLine;
  std::string sourceLine;
  std::string alignmentLine;
  size_t lineNum = 0;

  unsigned int numSentTried = 0;
  unsigned int numSentParsed = 0;

  while (true) {
    std::getline(targetStream, targetLine);
    std::getline(sourceStream, sourceLine);
    std::getline(alignmentStream, alignmentLine);
    lineNum++;

    if (targetStream.eof() && sourceStream.eof() && alignmentStream.eof()) {
      break;
    }

    if (targetStream.eof() || sourceStream.eof() || alignmentStream.eof()) {
      Error("Files must contain same number of lines");
    }

    std::vector<std::string> sourceTokens = ReadTokens(sourceLine);
    std::vector<std::string> targetTokens = ReadTokens(targetLine);

    // Read word alignments.
    Moses::Alignment alignment;
    try {
      alignment = Moses::ReadAlignment(alignmentLine);
    } catch (char const* e) {
      std::ostringstream s;
      s << "Failed to read alignment at line " << lineNum << ": " << e;
      Error(s.str());
    }
    if (alignment.size() == 0) {
      std::cerr << "skipping line " << lineNum << " without alignment points\n";
      continue;
    }

    std::cout << "##" << lineNum << "##" << std::endl;
    std::cout << sourceLine << std::endl;
    std::cout << targetLine << std::endl;

    TranslationUnitStore tuContainer(alignment, sourceTokens, targetTokens);
    //std::cout << tuContainer.toString();
    //tuStream << tuContainer.toStringTU();

    // filter long sentences
    if (sourceTokens.size() > options.maxLengthSource
        || targetTokens.size() > options.maxLengthTarget) {
      std::cout << "too long" << std::endl;
      continue;
    }
    // filter sentence not on the list
    if (!options.sentNumFile.empty()) {
      if (std::find(sentNumbers.begin(), sentNumbers.end(), lineNum) == sentNumbers.end()) {
        std::cout << "not on list" << std::endl;
        continue;
      }
    }

    numSentTried++;

    BottomUpHierarchicalAligner aligner(tuContainer,
        options.maxNumBlocksSource, options.maxNumBlocksTarget, options.normalForm);
    //aligner.printChart(); TODO
    //aligner.printGoalItems(); TODO
    aligner.printOneDerivation();
    if (aligner.success()) {
      numSentParsed++;
      outputStream << lineNum << "\t1" << std::endl;
    }
    else {
      outputStream << lineNum << "\t0" << std::endl;
    }

    // TODO
    //tuStream << tuContainer.toStringTUWithoutUnaligned(aligner.in2outSource, aligner.in2outTarget);

  }


  std::cout << "# sentences:        " << numSentTried << std::endl;
  std::cout << "# parsed sentences: " << numSentParsed << std::endl;

  PrintOutputFooter(outputStream, numSentTried, numSentParsed);

  return 0;
}

void ExtractLCFRS::OpenOutputFileOrDie(const std::string &filename,
    Moses::OutputFileStream &stream)
{
  bool ret = stream.Open(filename);
  if (!ret) {
    std::ostringstream msg;
    msg << "failed to open output file: " << filename;
    Error(msg.str());
  }
}

void ExtractLCFRS::OpenOutputFileOrDie(const std::string &filename,
                                      std::ofstream &stream)
{
  stream.open(filename.c_str());
  if (!stream) {
    std::ostringstream msg;
    msg << "failed to open output file: " << filename;
    Error(msg.str());
  }
}

void ExtractLCFRS::PrintOutputHeader(Moses::OutputFileStream& stream,
    int maxNumBlocksS, int maxNumBlocksT, int maxS, int maxT, bool nf)
{
  stream << "####################### Bottom-up hierarchical aligner #######################";
  stream << "#" << std::endl;
  stream << "#" << std::endl;
  stream << "# Maximal sentence length (source/target): " << maxS << "/" << maxT << std::endl;
  stream << "#" << std::endl;
  stream << "# Constraints:" << std::endl;
  stream << "#     Maximal number of blocks in source: " << maxNumBlocksS << std::endl;
  stream << "#     Maximal number of blocks in target: " << maxNumBlocksT << std::endl;
  stream << "#     Normal Form:                        " << (nf ? "yes" : "no") <<std::endl;
  stream << "#" << std::endl;
  stream << "###############################################################################" << std::endl;
}

void ExtractLCFRS::PrintOutputFooter(Moses::OutputFileStream& stream,
    int sent, int sentParsed)
{
  stream << "###############################################################################" << std::endl;
  stream << "# Number of sentences:        " << sent << std::endl;
  stream << "# Number of parsed sentences: " << sentParsed << std::endl;
  stream << "# Parse Failure Rate:         " << (sent-sentParsed)/(double)sent << std::endl;
  stream << "# Alignment Reachability:     " << sentParsed/(double)sent << std::endl;
}


void ExtractLCFRS::ProcessOptions(int argc, char *argv[],
                                 Options &options) const
{
  // inspired by moses/phrase-extract/extract-ghkm/ExtractGHKM::ProcessOptions
  namespace po = boost::program_options;
  namespace cls = boost::program_options::command_line_style;

  // Construct the 'top' of the usage message: the bit that comes before the
  // options list.
  std::ostringstream usageTop;
  usageTop << "Usage: " << GetName()
           << " [OPTION]* SOURCE TARGET ALIGNMENT OUTPUT\n"
           << "  SOURCE\tone-sentence per line source file\n"
           << "  TARGET\tone-sentence per line target file\n"
           << "  ALIGNMENT\tone-sentence per line alignment file: 0-0 0-1 2-5 etc.\n"
           << "  OUTPUT\toutput file name.\n\n"
           << "Alignment validation using a hierarchical bottom-up aligner.\n"
           << "Written by Miriam Kaeshammer (2013)\n\n"
           << "Options";

  // Declare the command line options that are visible to the user.
  int maxNumBlocks;
  int maxLength;
  po::options_description visible(usageTop.str());
  visible.add_options()
    ("noNF",
        "no restriction that grammar is in normal form")
    ("MaxNumBlocksSource",
        po::value(&options.maxNumBlocksSource)->default_value(options.maxNumBlocksSource),
        "set maximum number of blocks allowed per source item")
    ("MaxNumBlocksTarget",
        po::value(&options.maxNumBlocksTarget)->default_value(options.maxNumBlocksTarget),
        "set maximum number of blocks allowed per target item")
    ("MaxNumBlocks",
        po::value<int>(&maxNumBlocks),//po::value(&options.maxNumBlocks)->default_value(options.maxNumBlocks),
        "set maximum number of blocks allowed per monolingual item; overwrites the previous two options")
    ("MaxLengthSource",
        po::value(&options.maxLengthSource)->default_value(options.maxLengthSource),
        "set maximum number of tokens in the source sentence; longer sentences will be ignored")
    ("MaxLengthTarget",
        po::value(&options.maxLengthTarget)->default_value(options.maxLengthTarget),
        "set maximum number of tokens in the target sentence; longer sentences will be ignored")
    ("MaxLength",
       po::value<int>(&maxLength),
        "sets MaxLengthSource and MaxLengthTarget at the same time; overwrites the previous two options")
    ("SentenceNumbers",
        po::value(&options.sentNumFile),
        "only consider the sentences with numbers given in this file")
  ;

  // Declare the command line options that are hidden from the user
  // (these are used as positional options).
  po::options_description hidden("Hidden options");
  hidden.add_options()
    ("SourceFile",
        po::value(&options.sourceFile),
        "source file")
    ("TargetFile",
        po::value(&options.targetFile),
        "target file")
    ("AlignmentFile",
        po::value(&options.alignmentFile),
        "alignment file")
    ("OutputFile",
        po::value(&options.extractFile),
        "output file")
  ;

  // Compose the full set of command-line options.
  po::options_description cmdLineOptions;
  cmdLineOptions.add(visible).add(hidden);

  // Register the positional options.
  po::positional_options_description p;
  p.add("SourceFile", 1);
  p.add("TargetFile", 1);
  p.add("AlignmentFile", 1);
  p.add("OutputFile", 1);

  // Process the command-line.
  po::variables_map vm;
  const int optionStyle = cls::allow_long
                        | cls::long_allow_adjacent
                        | cls::long_allow_next;
  try {
    po::store(po::command_line_parser(argc, argv).style(optionStyle).
              options(cmdLineOptions).positional(p).run(), vm);
    po::notify(vm);
  } catch (const std::exception &e) {
    std::ostringstream msg;
    msg << e.what() << "\n\n" << visible;
    Error(msg.str());
  }

  if (vm.count("help")) {
    std::cout << visible << std::endl;
    std::exit(0);
  }

  // Check all positional options were given.
  if (!vm.count("TargetFile") ||
      !vm.count("SourceFile") ||
      !vm.count("AlignmentFile") ||
      !vm.count("OutputFile")) {
    std::ostringstream msg;
    std::cerr << visible << std::endl;
    std::exit(1);
  }

  // Process Boolean options.
  if (vm.count("noNF")) {
    options.normalForm = false;
  }

  // Adjust Options
  if (vm.count("MaxLength")) {
    options.maxLengthSource = maxLength;
    options.maxLengthTarget = maxLength;
  }
  if (vm.count("MaxNumBlocks")) {
    options.maxNumBlocksSource = maxNumBlocks;
    options.maxNumBlocksTarget = maxNumBlocks;
  }
}


void ExtractLCFRS::Error(const std::string &msg) const
{
  std::cerr << GetName() << ": " << msg << std::endl;
  std::exit(1);
}

std::vector<std::string> ExtractLCFRS::ReadTokens(const std::string &s)
{
  std::vector<std::string> tokens;
  std::string whitespace = " \t";

  std::string::size_type begin = s.find_first_not_of(whitespace);
  assert(begin != std::string::npos);
  while (true) {
    std::string::size_type end = s.find_first_of(whitespace, begin);
    std::string token;
    if (end == std::string::npos) {
      token = s.substr(begin);
    } else {
      token = s.substr(begin, end-begin);
    }
    tokens.push_back(token);
    if (end == std::string::npos) {
      break;
    }
    begin = s.find_first_not_of(whitespace, end);
    if (begin == std::string::npos) {
      break;
    }
  }

  return tokens;
}


}
